"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.StoreMeta = exports.Context = void 0;
const chalk_1 = __importDefault(require("chalk"));
const lodash_1 = __importDefault(require("lodash"));
const path = __importStar(require("path"));
const typescript_1 = __importDefault(require("typescript"));
const Formatting_1 = require("../Common/Formatting");
const Logger_1 = require("./Logger");
const { factory } = typescript_1.default;
class Context {
    constructor() {
        Object.defineProperty(this, "program", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "filePaths", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "modules", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "stores", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "config", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "compilerHost", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "printer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "typeChecker", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "parsedCommandLine", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
    }
    init(pluginConfig) {
        Context.instance = this;
        this.config = pluginConfig;
        const configPath = typescript_1.default.findConfigFile(pluginConfig.projectRoot, typescript_1.default.sys.fileExists, "tsconfig.json");
        if (!configPath) {
            throw new Error("Could not find a valid 'tsconfig.json'.");
        }
        const configFile = typescript_1.default.readConfigFile(configPath, typescript_1.default.sys.readFile);
        this.parsedCommandLine = typescript_1.default.parseJsonConfigFileContent(configFile.config, typescript_1.default.sys, "./");
        this.setFilePaths();
        this.compilerHost = typescript_1.default.createCompilerHost(this.parsedCommandLine.options);
        this.program = typescript_1.default.createProgram(this.filePaths, this.parsedCommandLine.options, this.compilerHost);
        this.printer = typescript_1.default.createPrinter({ newLine: typescript_1.default.NewLineKind.LineFeed });
        this.typeChecker = this.program.getTypeChecker();
        this.reloadModules();
    }
    setFilePaths() {
        const paths = this.config.storesDirectory.find({
            directories: false,
            files: true,
            recursive: true,
            matching: ['*Store.ts', '*store.ts'],
        }).map(filePath => typescript_1.default.sys.resolvePath(this.config.storesDirectory.path(filePath)));
        return this.filePaths = lodash_1.default.uniq(paths.concat(...this.parsedCommandLine.fileNames.map(f => typescript_1.default.sys.resolvePath(f))));
    }
    reloadModules() {
        this.setFilePaths();
        this.modules = this.program.getSourceFiles()
            .filter(module => this.filePaths.includes(module.fileName))
            .map(module => {
            return {
                module,
                store: new StoreMeta(module, this.config.storesPath)
            };
        });
    }
    process() {
        this.stores = [];
        for (const { module, store } of this.modules) {
            const outStore = this.processModule(module, store);
            if (!outStore)
                continue;
            this.stores.push(outStore.finalize());
        }
    }
    writeFiles() {
        var _a;
        const storeCount = (_a = this.stores) === null || _a === void 0 ? void 0 : _a.length;
        if (!storeCount) {
            (0, Logger_1.warnLog)('No stores located... skipping code-gen.');
            return;
        }
        (0, Logger_1.infoLog)('Located', chalk_1.default.reset(storeCount), 'stores:');
        (0, Logger_1.basicLog)(this.stores.map(s => "   - " + chalk_1.default.reset(s.className)).join('\n'));
        const vueDtsFile = createVueDtsFile(this.stores);
        const storesSource = typescript_1.default.createSourceFile(this.config.storesFileName, '', typescript_1.default.ScriptTarget.ES2022, false, typescript_1.default.ScriptKind.TS);
        const storesDtsTs = this.printer.printList(typescript_1.default.ListFormat.MultiLine, factory.createNodeArray(vueDtsFile), storesSource);
        this.config.storesDirectory.write(this.config.storesFileName, storesDtsTs);
        this.config.storesDirectory.write('StoreMeta.json', JSON.stringify(this.stores, null, 2));
        (0, Logger_1.successLog)('Generated stores vue declaration file at:', chalk_1.default.reset(path.join(this.config.storesPath, this.config.storesFileName).replace(this.config.projectRoot, '')));
    }
    rebuild() {
        this.process();
        this.writeFiles();
    }
    processModule(module, store) {
        var _a, _b, _c, _d, _e;
        for (let statement of module.statements) {
            // Search for our `export myStoreName = new MyStore();` statement and extract `myStoreName`
            if (typescript_1.default.isVariableStatement(statement)) {
                if (((_a = statement.modifiers) === null || _a === void 0 ? void 0 : _a.some(m => m.kind === typescript_1.default.SyntaxKind.ExportKeyword)) && ((_c = (_b = statement === null || statement === void 0 ? void 0 : statement.declarationList) === null || _b === void 0 ? void 0 : _b.declarations) === null || _c === void 0 ? void 0 : _c.length) === 1) {
                    const declaration = statement.declarationList.declarations[0];
                    if (typescript_1.default.isIdentifier(declaration.name)) {
                        store.exportName = declaration.name.text;
                    }
                }
            }
            // Search for our `export class MyStore extends Store<MyStore, IMyStoreState> {` statement and extract `MyStore`
            if (typescript_1.default.isClassDeclaration(statement)) {
                // Ensure our class extends Store
                const extendsStore = (_d = statement.heritageClauses) === null || _d === void 0 ? void 0 : _d.some(h => {
                    if (h.token !== typescript_1.default.SyntaxKind.ExtendsKeyword)
                        return false;
                    return h.types.some(t => {
                        var _a;
                        if (!typescript_1.default.isExpressionWithTypeArguments(t))
                            return false;
                        if (!typescript_1.default.isCallExpression(t.expression))
                            return false;
                        const callExpression = t.expression;
                        if (!typescript_1.default.isIdentifier(callExpression.expression))
                            return false;
                        return ((_a = callExpression.expression.text) === null || _a === void 0 ? void 0 : _a.toLowerCase()) === "store";
                    });
                });
                if (extendsStore && typescript_1.default.isIdentifier(statement.name)) {
                    store.className = statement.name.text;
                    // Now we have to look for the `public static vueBinding = 'x';` statement and pull out it's value.
                    for (let member of statement.members) {
                        if (typescript_1.default.isPropertyDeclaration(member)) {
                            if (!((_e = member.modifiers) === null || _e === void 0 ? void 0 : _e.some(m => m.kind === typescript_1.default.SyntaxKind.StaticKeyword)))
                                continue;
                            if (typescript_1.default.isIdentifier(member.name) && member.name.text === "vueBinding") {
                                if (typescript_1.default.isStringLiteral(member.initializer)) {
                                    store.vueBinding = member.initializer.text;
                                    break;
                                }
                            }
                        }
                    }
                    const actions = [];
                    for (let member of statement.members) {
                        if (typescript_1.default.isMethodDeclaration(member)) {
                            //							if (!member.modifiers?.some(m => m.kind === ts.SyntaxKind.PublicKeyword)) continue;
                            if (typescript_1.default.isIdentifier(member.name)) {
                                const action = new ActionMeta(member, member.name.text);
                                actions.push(action);
                                for (let param of member.parameters) {
                                    if (typescript_1.default.isIdentifier(param.name)) {
                                        action.addParam(param);
                                    }
                                }
                            }
                        }
                    }
                    store.actions = actions;
                }
            }
        }
        return store.isValid() ? store : undefined;
    }
    isStoreFile(path) {
        return this.stores.some(s => s.absFilePath === path);
    }
    getStoreByFilePath(path) {
        return this.stores.find(s => s.absFilePath === path);
    }
}
exports.Context = Context;
class StoreMeta {
    constructor(file, absStoresPath) {
        /**
         * The absolute path to the store file
         */
        Object.defineProperty(this, "absFilePath", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * The relative path of the store file, for ex, if importing in /src/stores/, and our file is at /src/stores/MyStore.ts, this will be MyStore.ts
         */
        Object.defineProperty(this, "relStoreFilePath", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * The name of the store file
         */
        Object.defineProperty(this, "name", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * The name of the class that extends Store
         */
        Object.defineProperty(this, "className", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * This will hold the name of our store export `export const store = new MyStore();`
         */
        Object.defineProperty(this, "exportName", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * If the dev defined `public static vueBinding = 'x';` in their store, this will hold that value
         */
        Object.defineProperty(this, "vueBinding", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "actions", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        this.absFilePath = file.fileName;
        this.name = path.basename(file.fileName);
        this.relStoreFilePath = path.relative(absStoresPath, file.fileName);
    }
    isValid() {
        return this.className !== undefined && this.exportName !== undefined;
    }
    finalize() {
        this.formatVueBinding();
        return this;
    }
    formatVueBinding() {
        if (this.vueBinding) {
            return;
        }
        /*let value = this.className;

         if (value.endsWith('Store')) {
         value = value.replace('Store', '');
         value = value.charAt(0).toLowerCase() + value.slice(1);
         }
         */
        this.vueBinding = (0, Formatting_1.formatVueBindingName)(this.vueBinding, this.className);
    }
}
exports.StoreMeta = StoreMeta;
class ActionMeta {
    constructor(method, name) {
        Object.defineProperty(this, "name", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "params", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "method", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "signature", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.method = method;
        this.name = name;
        this.signature = Context.instance.typeChecker.getSignatureFromDeclaration(this.method);
        for (let parameter of this.signature.parameters) {
            const ptype = Context.instance.typeChecker.typeToString(Context.instance.typeChecker.getTypeAtLocation(parameter.valueDeclaration));
            let defaultValue = undefined;
            if (typescript_1.default.isParameter(parameter.valueDeclaration)) {
                if (parameter.valueDeclaration.initializer) {
                    defaultValue = parameter.valueDeclaration.initializer.getText();
                }
            }
            this.params.push({
                name: parameter.name,
                type: ptype,
                defaultValue,
            });
        }
    }
    addParam(param) {
        //		const type = Context.instance.typeChecker.getSignatureFromDeclaration(this.method);
        //		this.params.push({
        //			name : (param.name as any).text,
        //			type : param.type?.getText()
        //		});
    }
    toJSON() {
        return {
            name: this.name,
            params: this.params
        };
    }
}
function createVueDtsFile(stores) {
    /**
     * This code will generate the following:
     *
     * declare module "@vue/runtime-core" {
     *   import {storeExportName} from "./StoreName";
     * 	 interface ComponentCustomProperties {
     * 		$storeVueBinding: typeof storeExportName;
     * 	 }
     * }
     *
     * export {};
     */
    /**
     * This will generate:
     * $storeVueBinding: typeof storeExportName;
     */
    const componentProperties = stores.map(store => factory.createPropertySignature(undefined, factory.createIdentifier(store.vueBinding), undefined, factory.createTypeQueryNode(factory.createIdentifier(store.exportName), undefined)));
    /**
     * This will generate our imports:
     * import {storeExportName} from "./StoreName";
     */
    const storeImports = stores.map(store => factory.createImportDeclaration(undefined, factory.createImportClause(false, undefined, factory.createNamedImports([
        factory.createImportSpecifier(false, undefined, factory.createIdentifier(store.exportName))
    ])), factory.createStringLiteral(`./${store.relStoreFilePath.replace('.ts', '')}`), undefined));
    /**
     * This will generate the main declare module/interface block
     */
    const declareModuleBlock = factory.createModuleDeclaration([factory.createModifier(typescript_1.default.SyntaxKind.DeclareKeyword)], factory.createStringLiteral("@vue/runtime-core"), factory.createModuleBlock([
        ...storeImports,
        factory.createInterfaceDeclaration(undefined, factory.createIdentifier("ComponentCustomProperties"), undefined, undefined, componentProperties)
    ]), typescript_1.default.NodeFlags.ExportContext /*| ts.NodeFlags.Ambient*/ | typescript_1.default.NodeFlags.ContextFlags);
    //	const imports = stores.map(store => `\timport {${store.exportName}} from './${store.relStoreFilePath.replace('.ts', '')}';`).join('\n');
    //	let contents = [
    //		`declare module "@vue/runtime-core" {`,
    //		imports,
    //		`\tinterface ComponentCustomProperties {`,
    //		stores.map(store => `\t\t${store.vueBinding}: typeof ${store.exportName};`).join('\n'),
    //		`\t}`,
    //		`}`,
    //		`export {}`
    //	];
    //	return contents.join('\n');
    return [
        declareModuleBlock,
        factory.createExportDeclaration(undefined, false, factory.createNamedExports([]), undefined, undefined),
    ];
}
//# sourceMappingURL=Generator.js.map